local factors        = require("scenario/gravitationalRacing/dataValues/factors")
local ClassPath      = require("scenario/gravitationalRacing/classes/classPath")
local ClassCelestial = require("scenario/gravitationalRacing/classes/classCelestial")

ClassBinarySystem = {}
ClassBinarySystem.__index = ClassBinarySystem

function ClassBinarySystem:getOrbitRadius(quantity, dist, mass, parentMass)
  --[[
  Returns the radius a celestial will orbit the barycenter at
  ]]--
  --r/[1 + (m1/m2)]
  if     quantity == "vector" then
    return dist:divide(1 + mass/parentMass)
  elseif quantity == "scalar" then
    return dist   /   (1 + mass/parentMass)
  end
end

function ClassBinarySystem:getOrbitalVelocity(component)
  --[[
  Returns the orbital velocity for this celestial
  ]]--
  local c1, c2
  if component == self.components[1] then
    c1 = component
    c2 = self.components[2]
  else
    c1 = self.components[1]
    c2 = component
  end

  local c2Mass = c2:getMass()
  local dist = c1:getPosition():getDistanceBetween(c2:getPosition()) / factors.getDistanceScaleFactor()

  --v = sqrt(G.m2.r1/r²)
  --If m1 = m2 -> v1 = v2 = sqrt(Gm/4r)
  return math.sqrt((ClassCelestial.getGravitationalConstant() * c2Mass * self:getOrbitRadius("scalar", dist, c1:getMass(), c2Mass)) / (dist^2)) * factors.getDistanceScaleFactor() * factors.getTimeScaleFactor()
end

function ClassBinarySystem:calculateBarycenter(components)
  --[[
  Calculates and return the barycenter of this system
  ]]--
  local c1, c2 = unpack(components)
  local c1Pos ,  c2Pos  = c1:getPosition(), c2:getPosition()
  local c1Mass,  c2Mass = c1:getMass()    , c2:getMass()

  --Needs to be absolute since to work out the barycenter, the distance vector must be positive
  local distVec = c1Pos:subtract(c2Pos):abs()
  --This is relative to the components
  local barycenterRelative = self:getOrbitRadius("vector", distVec, c1Mass, c2Mass)

  --Subtract from the one in front (could also add to the one behind)
  local celestialInFront = c1Pos:getY() > c2Pos:getY() and c1Pos or c2Pos
  return celestialInFront:subtract(barycenterRelative)
end

function ClassBinarySystem:new(systemName, components)
  local self = {}

  --A system can be orbited by another star or system so can have its own path
  self.systemPath = nil
  self.systemName = systemName
  self.systemMass = 0

  self.components = components
  self.barycenter = ClassVector.new(0, 0, 0)

  setmetatable(self, ClassBinarySystem)
  return self
end

function ClassBinarySystem:initialise()
  --[[
  Initialises some attributes that need to be set after the constructor or components occurs
  ]]--
  self.systemMass = self.components[1]:getMass() + self.components[2]:getMass()
  self.barycenter = ClassBinarySystem:calculateBarycenter(self.components)
end

function ClassBinarySystem:calculateFrequency()
  --[[
  Returns the frequency of the system
  ]]--
  local c1, c2 = unpack(self.components)
  --T = sqrt(4pi² r³ / G(m + M))
  local period = math.sqrt(4*math.pi^2 * (c1:getPosition():getDistanceBetween(c2:getPosition()) / factors.getDistanceScaleFactor())^3 / (ClassCelestial.getGravitationalConstant() * self.systemMass)) / factors.getTimeScaleFactor()
  return 1/period
end

function ClassBinarySystem:setupComponentPaths()
  --[[
  Calculates and configures the paths for both celestials in the system
  ]]--
  local p1, p2 = unpack(self.components)

  --Don't setup paths for delete celestials
  if not (p1:isRemoved() or p2:isRemoved()) and not (p1:getPath() or p2:getPath()) then
    local data = {}
    for i, instance in ipairs(self.components) do
      data[i] = {
        mass = instance:getMass(),
        name = instance:getName(),
        pos  = instance:getPosition()
      }
    end

    local dist = data[1].pos:getDistanceBetween(data[2].pos)

    local p1OrbitRadius = self:getOrbitRadius("scalar", dist, data[1].mass, data[2].mass)
    local p2OrbitRadius = self:getOrbitRadius("scalar", dist, data[2].mass, data[1].mass)

    local config = {
      frequency       = nil,
      invertDirection = nil
    }

    --Setup up config values
    for i = 1, 2 do
      local obj = scenetree.findObject(self.components[i]:getName())
      for k, v in pairs(config) do
        --If two objects have different values, take the first's
        config[k] = v or obj[k]
      end
    end

    if not config.frequency then
      config.frequency = self:calculateFrequency()
    end

    --Adjust names to be valid for torquescript
    local p1Path = ClassPath.newCircularPath(data[1].name:gsub("[(|)]", {["("] = "_", ["|"] = "_", [")"] = "_"}), 0      , p1, self.barycenter, p1OrbitRadius, config.invertDirection, config.frequency)
    local p2Path = ClassPath.newCircularPath(data[2].name:gsub("[(|)]", {["("] = "_", ["|"] = "_", [")"] = "_"}), math.pi, p2, self.barycenter, p2OrbitRadius, config.invertDirection, config.frequency)
    p1:setPath(p1Path)
    p2:setPath(p2Path)

    --Show the second trail if it will be different to the first
    --Allow for two different stars orbitting but having mas defects due to rounding/floating point arithmetic
    if math.abs(p1OrbitRadius - p2OrbitRadius) > 1 then
      p2Path:createTrail(true)
    end

    p1Path:createTrail(true)
  end
end

function ClassBinarySystem:getCelestialColour()
  --[[
  Returns the colour of the system
  ]]--
  local c1, c2 = self.components[1]:getCelestialColour(), self.components[2]:getCelestialColour()
  --Return the average of both rgb values
  return {
    r = (c1.r + c2.r)/2,
    g = (c1.g + c2.g)/2,
    b = (c1.b + c2.b)/2
  }
end

function ClassBinarySystem:setPath(path)
  --[[
  Setups the system path
  ]]--
  if not self.systemPath then
    self.systemPath = path
  end
end

function ClassBinarySystem:getPath()
  return self.systemPath
end

function ClassBinarySystem:getComponents()
  return self.components
end

function ClassBinarySystem:isRemoved()
  return self.components[1]:isRemoved() and self.components[2]:isRemoved()
end

function ClassBinarySystem:getMass()
  return self.systemMass
end

function ClassBinarySystem:getPosition()
  return self.barycenter
end

function ClassBinarySystem:getName()
  return self.systemName
end

function ClassBinarySystem:update(dt)
  if self.systemPath then
    --If both components don't exist, stop updating the system
    if not (self.components[1]:isRemoved() and self.components[2]:isRemoved()) then
      self.barycenter = self.systemPath:update(dt)

      for _, component in ipairs(self.components) do
        component:getPath():setOffset(self.barycenter)
      end
    end
  end
end

function ClassBinarySystem:instanceOf()
  return "ClassBinarySystem"
end

-----------------------------------------------------------------------------------------------------------------------------------------------

local function new(systemName, components)
  return ClassBinarySystem:new(systemName, components)
end

local M = {}

M.new = new
return M
